/* BlendingMode.h
Copyright (c) 2024 by tibetiroka

Endless Sky is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

Endless Sky is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
*/

#pragma once



enum class BlendingMode : char {
	ALPHA = '-',
	PREMULTIPLIED_ALPHA = '=',
	HALF_ADDITIVE = '^',
	ADDITIVE = '+',
	COMPAT_HALF_ADDITIVE = '~' // The legacy half-additive mode marker. This is only used during parsing.
};
